# Version 0.0.3
# Thomas Lachmann 2013 <blender@automatic-brain.de>

# original file Tamito KAJIYAMA <19 August 2009>
# - DXF output instead of SVG
# - color mapping to ACI colors
# - rendersize calculated
# - frustum culling added
# 2013-12-10
# - parameters from dxfWRiterGUI.py added for controlling the generated edges
# 2013-12-14
#  - no use of dxfLibrary any more
#  - write dxf R14 files with truecolor support
#  - implemented a color shader which takes the original material color

from freestyle import *
from logical_operators import *
from shaders import *
import bpy
import os
import freestyle
import math
from mathutils import *
from parameter_editor import *
from dxfFile import *

CLIP_LEFT = 1
CLIP_RIGHT = 2
CLIP_TOP = 3
CLIP_BOTTOM = 4

class myMaterialColorShader(StrokeShader):
	def __init__(self):
		StrokeShader.__init__(self)
	def shade(self, stroke):
		it = stroke.stroke_vertices_begin()
		func = MaterialF0D()
		while not it.is_end:
			mat = func(Interface0DIterator(it))
			
			r = mat.diffuse[0]
			g = mat.diffuse[1]
			b = mat.diffuse[2]

			it.object.attribute.color = (r, g, b)
			it.increment()


class DXFWriter(StrokeShader):
	def __init__(self, f, w, h):
		StrokeShader.__init__(self)
		self.width, self.height = w, h
		# create the file 	
		self.file = f
		self.file.write(DXF_HEADER % (w, h))
		self.dxfHandle = 100
		
	def close(self):
		try:
			self.file.write(DXF_FOOTER)
			self.file.close()
			print("dxf saved to file %r" % self.file.name)
		except:
			print("Cannot save to file %r" % self.file.name)

	def isPointOutsideViewport(self, point): 
		x, y = point
		if x < 0:
			return CLIP_LEFT
		if y < 0:
			return CLIP_BOTTOM
		if x > self.width:
			return CLIP_RIGHT
		if y > self.height:
			return CLIP_TOP
		return None
			
	def shade(self, stroke):

		points = []
		for v in stroke:
			x, y = v.point
			points.append((x, y))
		
		points_clipped = []			
		if len(stroke) > 0:
			last_clip = self.isPointOutsideViewport(points[0])
			if last_clip == None:
				points_clipped.append(DXF_VERTEX % (points[0][0], points[0][1]))

		for i in range(1, len(points)):
			clip = self.isPointOutsideViewport(points[i])
			if (clip == None) and (last_clip == None):
				points_clipped.append(DXF_VERTEX % (points[i][0], points[i][1]))
			else:
				clip_tmp = last_clip
				if (last_clip == None):
					clip_tmp = clip
				if clip_tmp == CLIP_LEFT:
					clip_p1 = (0,0)
					clip_p2 = (0,self.height)
				elif clip_tmp == CLIP_RIGHT:
					clip_p1 = (self.width,0)
					clip_p2 = (self.width,self.height)
				elif clip_tmp == CLIP_TOP:
					clip_p1 = (0,self.height)
					clip_p2 = (self.width,self.height)
				elif clip_tmp == CLIP_BOTTOM:
					clip_p1 = (0,0)
					clip_p2 = (self.width,0)
				intersect = mathutils.geometry.intersect_line_line_2d(mathutils.Vector(clip_p1), mathutils.Vector(clip_p2), mathutils.Vector(points[i-1]), mathutils.Vector(points[i]))
				if intersect:
					points_clipped.append(DXF_VERTEX % (intersect.x, intersect.y))
					if last_clip != None:
						points_clipped.append(DXF_VERTEX % (points[i][0], points[i][1]))
				last_clip = clip
		
		if len(points_clipped) > 0:
			self.file.write(DXF_LWPOLYLINE % (self.dxfHandle, dxfColor(v.attribute.color), len(points_clipped), "".join(points_clipped)))
			self.dxfHandle += 1


scene = freestyle.getCurrentScene()

output_dir = bpy.path.abspath(scene.render.filepath)
if not os.path.exists(output_dir):
	os.makedirs(output_dir)
path = os.path.join(output_dir, "Frame%06d.dxf" % scene.frame_current)
f = open(path, "wt")

w = scene.render.resolution_x * scene.render.resolution_percentage / 100
h = scene.render.resolution_y * scene.render.resolution_percentage / 100

if  hasattr(scene, 'dxfWriter_settings'):
    print('dxfWriter settings found.')
    lineset = scene.dxfWriter_settings
    selection_criteria = []
    # prepare selection criteria by visibility
    if lineset.select_by_visibility:
        if lineset.visibility == 'VISIBLE':
            selection_criteria.append(
                QuantitativeInvisibilityUP1D(0))
        elif lineset.visibility == 'HIDDEN':
            selection_criteria.append(
                NotUP1D(QuantitativeInvisibilityUP1D(0)))
        elif lineset.visibility == 'RANGE':
            selection_criteria.append(
                QuantitativeInvisibilityRangeUP1D(lineset.qi_start, lineset.qi_end))
    # prepare selection criteria by edge types
    if lineset.select_by_edge_types:
        edge_type_criteria = []
        if lineset.select_silhouette:
            upred = pyNatureUP1D(Nature.SILHOUETTE)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_silhouette else upred)
        if lineset.select_border:
            upred = pyNatureUP1D(Nature.BORDER)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_border else upred)
        if lineset.select_crease:
            upred = pyNatureUP1D(Nature.CREASE)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_crease else upred)
        if lineset.select_ridge_valley:
            upred = pyNatureUP1D(Nature.RIDGE)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_ridge_valley else upred)
        if lineset.select_suggestive_contour:
            upred = pyNatureUP1D(Nature.SUGGESTIVE_CONTOUR)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_suggestive_contour else upred)
        if lineset.select_material_boundary:
            upred = pyNatureUP1D(Nature.MATERIAL_BOUNDARY)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_material_boundary else upred)
        if lineset.select_edge_mark:
            upred = pyNatureUP1D(Nature.EDGE_MARK)
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_edge_mark else upred)
        if lineset.select_contour:
            upred = ContourUP1D()
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_contour else upred)
        if lineset.select_external_contour:
            upred = ExternalContourUP1D()
            edge_type_criteria.append(NotUP1D(upred) if lineset.exclude_external_contour else upred)
        if lineset.edge_type_combination == 'OR':
            upred = join_unary_predicates(edge_type_criteria, OrUP1D)
        else:
            upred = join_unary_predicates(edge_type_criteria, AndUP1D)
        if upred is not None:
            if lineset.edge_type_negation == 'EXCLUSIVE':
                upred = NotUP1D(upred)
            selection_criteria.append(upred)
    # prepare selection criteria by face marks
    if lineset.select_by_face_marks:
        if lineset.face_mark_condition == 'BOTH':
            upred = FaceMarkBothUP1D()
        else:
            upred = FaceMarkOneUP1D()
        if lineset.face_mark_negation == 'EXCLUSIVE':
            upred = NotUP1D(upred)
        selection_criteria.append(upred)
    # prepare selection criteria by group of objects
    #~ if lineset.select_by_group:
        #~ if lineset.group is not None:
            #~ names = dict((ob.name, True) for ob in lineset.group.objects)
            #~ upred = ObjectNamesUP1D(names, lineset.group_negation == 'EXCLUSIVE')
            #~ selection_criteria.append(upred)
    # prepare selection criteria by image border
    if lineset.select_by_image_border:
        xmin, ymin, xmax, ymax = ContextFunctions.get_border()
        upred = WithinImageBoundaryUP1D(xmin, ymin, xmax, ymax)
        selection_criteria.append(upred)
    # select feature edges
    upred = join_unary_predicates(selection_criteria, AndUP1D)
    if upred is None:
        upred = TrueUP1D()
else:
    upred = QuantitativeInvisibilityUP1D(0)
    
Operators.select(upred)

#Operators.bidirectionalChain(ChainSilhouetteIterator(), NotUP1D(upred))

writer = DXFWriter(f, w, h)
shaders_list = [
	ConstantThicknessShader(2),
	myMaterialColorShader(),
	writer,
]
Operators.create(TrueUP1D(), shaders_list)
writer.close()

